import engine
import data
import tile
import rpg
import dialogs

import math
import pickle
import pygame
from pygame.constants import *

class Cursor(rpg.TileActor):
    def __init__(self, tilemap):
        rpg.TileActor.__init__(self, tilemap)
        self.yellow = data.load_image("cursor_yellow.png")
        self.red = data.load_image("cursor_red.png")
        self.yellow.set_colorkey( (255, 0, 255) )
        self.red.set_colorkey( (255, 0, 255) )
        self.image = self.yellow
        self._world.size = (self.yellow.get_width(), self.yellow.get_height())
        self._rect.size = self._world.size
        self.translateable = False
        self.blocking = False
        
    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['red']
        del odict['yellow']
        del odict['image']
        return odict
    
    def __setstate__(self,dict):        
        self.__dict__.update(dict)
        self.yellow = data.load_image("cursor_yellow.png")
        self.red = data.load_image("cursor_red.png")
        self.image = self.yellow
        
    def mouseMove(self, pos):
        self._world.center = self.tilemap.screenToWorld(pos)
        
    def click(self, but, at):
        # We're assuming all clicks on the screen are for us.
        potential = self.tilemap.occupiedBy(self.tLoc, [self.tilemap.center])
        if potential:
            self.tilemap.center.target = potential[0]
        
    def paint(self, screen):
        screen.blit(self.image, self._rect)
        
    def update(self, delay):
        self.tLoc = self.tilemap.worldToTile(self._world.center)
        self._rect.center = self.tilemap.tileToScreen(self.tLoc)
        if(self.tilemap.isOccupied(self.tLoc, [self.tilemap.center])):
            self.image = self.red
        else:
            self.image = self.yellow
        
class LogHud(engine.Box):
    def __init__(self, pgs):
        self.pgs = pgs
        y = 3
        x = 3
        engine.Box.__init__(self, pygame.Rect(x, y, pgs.driver.screen.w-6, 0))
        self.messages = []
        self.noNews = 0
        
    def postMessage(self, message):
        self.noNews = 0
        label = engine.Label(message)
        self.add(label)
        self.messages.append(label)
        if len(self.messages) > 5:
            self.remove(self.messages.pop(0))
        self.layoutMessages()
        
    def layoutMessages(self):
        top = self._rect.top + 3
        left = self._rect.left + 3
        
        for l in self.messages:
            l.rect.top = top
            l.rect.left = left
            top += l.rect.height + 3
        
        self._rect.height = top - self._rect.top
        
    def update(self, delay):
        engine.Box.update(self,delay)
        self.noNews += delay
        if self.noNews >= 1500:
            self.noNews = 0
            if self.messages: 
                self.remove(self.messages.pop(0))
                self.layoutMessages()
        
class PlayerHud(engine.Box):
    def __init__(self, pgs):
        self.pgs = pgs
        y = pgs.driver.screen.h - (pgs.driver.screen.h / 5)
        engine.Box.__init__(self, pygame.Rect( 0, y, pgs.driver.screen.w,
                                              pgs.driver.screen.h / 5))
        self.nameLabel = engine.Label("Player: " + pgs.pc.name)
        self.nameLabel.rect.topleft = ( self._rect.x + 3, self._rect.y + 3)
        
        self.add(self.nameLabel)
        
        self.hpLabel = engine.Label("Level: x HP: y/z ")
        loc = self.nameLabel.rect.bottomleft
        self.hpLabel.rect.topleft = (loc[0], loc[1] + 3)
        self.add(self.hpLabel)
        
        self.targetLabel = engine.Label("No Target")
        self.targetLabel.rect.topleft = (self._rect.w / 2, self._rect.y + 3)
        self.add(self.targetLabel)
        
    def update(self, delay):
        engine.Box.update(self,delay)
        pc = self.pgs.pc
        dataTuple = (pc.level, pc.hp, pc.maxHp)
        self.hpLabel.text = "Level: %d HP: %d / %d" % dataTuple
        
        if pc.target:
            target = pc.target
            dataTuple = (target.name, target.hp, target.maxHp)
            self.targetLabel.text = "Target: %s (%d/%d)" % dataTuple
        else:
            self.targetLabel.text = "No Target"

class HotkeyHud(engine.Box):
    def __init__(self, pgs):
        engine.Box.__init__(self, pygame.Rect(3, 175,0,0))
        self.pgs = pgs
        
        x = self._rect.x + 3
        y = self._rect.y + 3
        w = 0
        
        self.hotkeyLabels = []
        for index in range(10):
            number = str( (index + 1) % 10 )
            label = engine.Button("<placeholder>", self.use,number)
            label.rect.topleft = [ x, y ]
            if label.rect.w > w: w = label.rect.w
            y += label.rect.h + 3
            self.add(label)
            self.hotkeyLabels.append(label)
        
        self._rect.size = ( w, y - self._rect.top )
        
    def update(self, delay):
        engine.Box.update(self,delay)
        
        w = self._rect.w
        for x in range(10):
            skill = self.pgs.hotkeys[x]
            text = "%d: " % ((x + 1) % 10)
            label = self.hotkeyLabels[x]
            if skill:
                text += skill.name
            if label.text != text:
                label.text = text
            if label.rect.w +6 > w: w = label.rect.w + 6
        self._rect.w = w 
        
    def use(self, button):
        index = self.hotkeyLabels.index(button)
        skill = self.pgs.hotkeys[index]
        if skill:
            skill.apply(self.pgs.pc)

class PlayingGameState(engine.State):
    def __init__(self, driver, pc):
        engine.State.__init__(self,driver)
        self.pc = pc
        self.hotkeys = [None for x in range(10) ]
        self.tiles = None
        self.level = 0
        self.cfts = None
        self.save = None
        
        self.playerHud = PlayerHud(self)
        self.keyHud = HotkeyHud(self)
        self.log = LogHud(self)
        self.setupTiles()
    
        self.log.postMessage("Welcome to the Tower of Trials!")
        
    def activate(self):
        if self.cfts:
            chosen = self.cfts.chosen
            if(chosen != None):
                chosenLoot = self.sortedLoot[chosen]
                self.tiles.takeLoot(chosenLoot)
                self.pc.inventory.append(chosenLoot.item)
            self.cfts = None
        elif self.save:
            self.save = None
            outfile = file(self.pc.name+'.sav','wb')
            pickle.dump(self,outfile)
            outfile.close()
            raise SystemExit
        
    def setupTiles(self):
        self.spawnCounter = 0
        if self.tiles:
            self.remove(self.tiles)
        self.tiles = tile.TileMap(tile.RandomSource(30,30), 
                                  self.driver.screen.size,self.log)
        self.pc.tilemap = self.tiles
        
        cursor = Cursor(self.tiles)
        self.tiles.add(cursor)

        stairs = rpg.Stairs(self.tiles)
        self.tiles.addStairs(stairs)
        
        self.tiles.add(self.pc)
        self.tiles.center = self.pc
    
        self.add(self.tiles)
        
        self.remove(self.playerHud)
        self.add(self.playerHud)
        self.remove(self.keyHud)
        self.add(self.keyHud)
        self.remove(self.log)
        self.add(self.log)
        
    def keyEvent(self, event):
        key = event.key
        pressed = event.type == KEYDOWN
        if(key == K_i and pressed):
            inventory = dialogs.InventoryState(self)
            self.driver.push(inventory)
        elif(key == K_c and pressed):
            pcLoc = self.tiles.worldToTile(self.pc.world)
            stairLoc = self.tiles.stairLoc()
            dist = math.sqrt(math.pow(pcLoc[0] - stairLoc[0],2) +
                             math.pow(pcLoc[1] - stairLoc[1],2))
            # God-damned hack because stairs aren't where they appear to be.
            if dist <= 2.0:
                self.level += 1
                self.pc.target = None
                self.setupTiles()
        elif(key == K_g and pressed):
            loot = self.tiles.loot
            comparator =(lambda loot1,loot2:
                cmp(data.distance(loot1.world.center, self.pc.world.center),
                    data.distance(loot2.world.center, self.pc.world.center)))
            loot.sort(comparator)
            self.sortedLoot = loot
            self.cfts = dialogs.ChooseFromTenState(self,loot[:10],"Loot what?")
            self.driver.push(self.cfts)
        elif(key == K_s and pressed):
            self.save = dialogs.ConfirmDialog(self.driver, "Save and quit?")
            self.driver.push(self.save)
            
        engine.State.keyEvent(self,event)
        
    def spawnEnemy(self):
        level = int(1 + self.level/2)
        skull = rpg.Enemy("skull.png", self.tiles,10,level)
        skull.world.center = self.tiles.tileToWorld(self.tiles.randomOpenTile())
        skull.name = "Skull"
        self.tiles.add(skull)
        
    def update(self, delay):
        self.spawnCounter += delay
        if self.spawnCounter >= 10000:
            self.spawnEnemy()
            self.spawnCounter = 0
        engine.State.update(self,delay)
        if self.pc.isDead():
            over = dialogs.GameOverState(self)
            self.driver.pop()
            self.driver.push(over)
        