import pygame
import math
import random
from pygame.constants import *

import engine
import data

def formatBonus(bonus):
    if bonus > 0:
        return "+%d" % bonus
    elif bonus == 0:
        return ""
    else:
        return "%d" % bonus

class LevelProgression(object):
    
    STATS = [ "hp", "base", "fort", "ref", "will" ]
    
    def __init__(self):
        self.name="None"
    
    def getXpForLevel(self, level):
        if level == 1: return 0
        return getXpForLevel(level-1) + (level-1) * 1000
    
    def levelup(self, levelingTo):
        raise NotImplementedError
    
class WarriorLevelProgression(LevelProgression):
    def __init__(self):
        LevelProgression.__init__(self)
        self.name="Warrior"
        
    def levelup(self, levelingTo):
        return dict(zip(self.STATS, (10, 1, 0, 0, 0)))

class Item(object):
    def __init__(self, name):
        self.equippable = False
        self.useable = False
        self.stackable = False
        self.name = name
        
class Skill(object):
    def __init__(self, name):
        self.name = name
        
    def apply(self, pc):
        raise NotImplementedError
        
class Potion(Item, Skill):
    def __init__(self, name):
        Item.__init__(self, name)
        Skill.__init__(self,name)
        self.useable = True
        self.stackable = True
        self.amount = 1
        
    def __str__(self):
        return "%s (x%d)" % (self.name, self.amount)
 
class Equippable(Item):
    ARMOR = 0
    WEAPON = 1
    def __init__(self, name, type):
        Item.__init__(self,name)
        self.equippable = True
        self.type = type
    
class Armor(Equippable):
    TORSO = 'Torso'
    
    def __init__(self, name, slot, ac, bonus=0):
        Equippable.__init__(self, name, Equippable.ARMOR)
        self.bonus = bonus
        self.ac = ac
        self.slot = slot
        
    def getTotalAC(self):
        return self.ac + self.bonus
    
    def __str__(self):
        result = "%s %s (%d AC)" % (formatBonus(self.bonus), self.name,
                                    self.getTotalAC())
        return result
    
class Weapon(Equippable):
    def __init__(self, name, dx, bonus = 0):
        Equippable.__init__(self, name, Equippable.WEAPON)
        self.bonus = bonus
        self.dx = dx
        
    def damageRange(self):
        return (1, self.dx)
    
    def getBonus(self):
        return self.bonus
    
    def rollDamage(self):
        return random.randint(1, self.dx)
    
    def __str__(self):
        return "%s %s" % (formatBonus(self.bonus), self.name)

class TileActor(engine.Actor):
    def __init__(self, tilemap):
        engine.Actor.__init__(self)
        self._tilemap = tilemap
        self._world = pygame.Rect(0,0,0,0)
        self.filename = None
        self.image = None
        self.translateable = True # Whether or not the map should set rect for
                                  # us
        self.blocking = True # Whether or not we impede movement
    
    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['image']
        return odict
    
    def __setstate__(self, dict):
        img = data.load_image(dict['filename'])
        img.set_colorkey(img.get_at( (0,0) ))
        dict['image'] = img
        
        self.__dict__.update(dict)
    
    def getTilemap(self):
        return self._tilemap
    
    def isDead(self):
        return False
        
    def setTilemap(self, tilemap):
        "Only the PC should actually call this"
        self._tilemap = tilemap
        tLoc = tilemap.randomOpenTile()
        self._world.center = tilemap.tileToWorld(tLoc)
        
    def getWorld(self):
        return self._world

    tilemap = property(getTilemap, setTilemap)
    world = property(getWorld)

class Stairs(TileActor):
    def __init__(self, tilemap):
        TileActor.__init__(self, tilemap)
        self.blocking = False
        self.filename = "stairs.png"
        self.image = data.load_image("stairs.png")
        size = (self.image.get_width(),self.image.get_height())
        self.world.size = size
        self.rect.size = size
        
    def paint(self, screen):
        screen.blit(self.image, self._rect)
        
class Loot(TileActor):
    def __init__(self, tilemap, item, filename="sack.png"):
        TileActor.__init__(self, tilemap)
        self.blocking = False
        self.item = item
        self.filename = filename
        self.image = data.load_image(filename)
        self.image.set_colorkey(self.image.get_at( (0,0)))
        size = (self.image.get_width(),self.image.get_height())
        self.world.size = size
        self.rect.size = size
        self.name = item.name
        
    def paint(self, screen):
        screen.blit(self.image, self._rect)

class ReinerActor(TileActor):
    DIRS = 8
    MS_PER_FRAME = 100
    
    FACING_TO_DIRS = [ (0, -1), (1, -1), (1, 0),
                       (1,  1), (0,  1), (-1, 1),
                       (-1, 0), (-1, -1) ]
    
    FACING_TO_DIRS = dict(enumerate(FACING_TO_DIRS))
    DIRS_TO_FACING = dict(zip(FACING_TO_DIRS.values(),
                              FACING_TO_DIRS.keys()))
    
    def __init__(self, filename, tilemap, frames = 8):
        TileActor.__init__(self,tilemap)
        self.filename = filename
        self.image = data.load_image(filename)
        self.image.set_colorkey(self.image.get_at( (0,0) ))
        self.frames = frames
        self.facing = 0
        self.frame = 0
        self.time = 0
        self.moving = 0
        self.attacking = False
        self._target = None
        self.name = "Mr. Uninitialized"
        self.hp = 0
        self.maxHp = 0
        self.level = 0
        self.base_attack = 0
        self.attack_speed = 1000 # ms / attack
        self.attack_buildup = 0
        self.speed = 150 / 1000.0 # in pixels/ms
        self.inventory = []
        self.weapon = None
        self.armor = {}
        self.progression = WarriorLevelProgression()
        self.charWidth = self.image.get_width() / self.frames
        self.charHeight= self.image.get_height()/ self.DIRS
        self._world.size = (self.charWidth, self.charHeight)
        self._rect = pygame.Rect(self._world)
        
    def attack(self):
        toHit = random.randint(1,20)
        # Ignoring crits for now
        # Ignoring melee vs. ranged for now
        if(toHit + self.melee > self.target.ac):
            data = (self.name, self.target.name, self.damage)
            self.tilemap.log.postMessage("%s hit %s for %d damage!" % data)
            self.target.takeDamage(self, self.damage)
        else:
            self.tilemap.log.postMessage("%s missed %s!" % (self.name,
                                                            self.target.name))
            self.target.missed(self)
        
    def deequip(self, item):
        if item == self.weapon:
            self.weapon = None
        else:
            for slot in self.armor.keys():
                armor = self.armor[slot]
                if item == armor:
                    del self.armor[slot]
                    break
        
    def doUpdate(self, delay):
        advance = 0
        self.time += delay
        if self.time > self.MS_PER_FRAME:
            self.time -= self.MS_PER_FRAME
            advance = 1
        if self.moving:
            dx,dy = self.FACING_TO_DIRS[self.facing]
            self._world.x += dx * self.speed * delay
            self._world.y += dy * self.speed * delay
            if advance:
                self.frame += 1
                self.frame %= self.frames
        
    def equip(self, equippable):
        
        if not equippable in self.inventory:
            self.inventory.append(equippable)
        if equippable.equippable:
            if equippable.type == Equippable.WEAPON:
                self.weapon = equippable
            elif equippable.type == Equippable.ARMOR:
                self.armor[equippable.slot] = equippable
        
    def getAC(self):
        armor = 0
        for slot in self.armor.keys():
            item = self.armor[slot]
            armor += item.getTotalAC()
        return 10 + armor
    
    def getDamage(self):
        # For now, everyone does a d4 of damage
        return random.randint(1,4)
        
    def getMeleeBonus(self):
        wpn = 0
        if self.weapon:
            wpn = self.weapon.bonus
        return self.base_attack + wpn
        
    def getTarget(self):
        return self._target
    
    def heal(self, amount):
        self.hp += amount
        if self.hp > self.maxHp: self.hp = self.maxHp
    
    def isDead(self):
        return self.hp < 0
    
    def levelup(self):
        self.level += 1
        progress = self.progression.levelup(self.level)
        self.maxHp += progress['hp']
        self.base_attack += progress['base']
        
        self.hp = self.maxHp
    
    def missed(self, attacker):
        pass
    
    def paint(self, screen):
        source = pygame.Rect( self.frame * self.charWidth, 
                              self.facing * self.charHeight,
                              self.charWidth,
                              self.charHeight )
        screen.blit(self.image, self._rect, source)
                
    def setTarget(self, target):
        if self._target != target: self.attack_buildup = 0
        self._target = target
        
    def takeDamage(self, source, amount):
        self.hp -= amount
        
    def tryUpdate(self, delay):
        # With the player, we have to make sure they won't hit the walls
        oldLoc = self._world.center
        
        # Used to be midbottom
        displacedBottom = self._world.midbottom[1] - 20
        midbottom = (self._world.midbottom[0], displacedBottom)
        
        olddisplaced = midbottom
        oldTileCenter = self.tilemap.worldToTile(olddisplaced)
        
        self.doUpdate(delay)
        
        displacedBottom = self._world.midbottom[1] - 20
        midbottom = (self._world.midbottom[0], displacedBottom)
        newTileCenter = self.tilemap.worldToTile(midbottom)
        
        if(newTileCenter != oldTileCenter):
            if(not self.tilemap.source.isWalkable(newTileCenter) or
               self.tilemap.isOccupied(newTileCenter, [self])):
                self._world.center = oldLoc
        
    def update(self, delay):
        self.tryUpdate(delay)
        
        if self._target and self.attacking:
            self.attack_buildup += delay
            if self.attack_buildup > self.attack_speed:
                self.attack_buildup -= self.attack_speed
                
                self.attack()

    ac = property(getAC)
    damage = property(getDamage)
    melee = property(getMeleeBonus)
    target = property(getTarget, setTarget)
    
class PC(ReinerActor):
    def __init__(self, filename):
        ReinerActor.__init__(self,filename,None)
        self.dx = 0
        self.dy = 0
        self.autoAttack = True
        
    def keyEvent(self, event):
        pressed = 1
        if event.type == KEYUP: pressed = 0
        
        dx = self.dx
        dy = self.dy
        
        if event.key == K_UP:
            dy = -pressed
        elif event.key == K_RIGHT:
            dx = pressed
        elif event.key == K_DOWN:
            dy = pressed
        elif event.key == K_LEFT:
            dx = -pressed
            
        self.dx = dx
        self.dy = dy
        self.moving = dx or dy
        if self.moving:
            self.facing = self.DIRS_TO_FACING[ (dx,dy) ]
            
    def missed(self, attacker):
        self.retaliate(attacker)
    
    def retaliate(self, target):
        if (((not self.target) or self.target == target) and self.autoAttack):
            self.target = target
            self.attacking = True
    
    def takeDamage(self, attacker, amount):
        ReinerActor.takeDamage(self,attacker, amount)
        self.retaliate(attacker)
                
class NPC(ReinerActor):
    def __init__(self, filename, tilemap, frames):
        ReinerActor.__init__(self,filename,tilemap, frames)
        self.facing = 4
        
class Enemy(NPC):
    def __init__(self, filename, tilemap, frames, level):
        NPC.__init__(self, filename, tilemap, frames)
        for x in range(level): self.levelup()
        self.pc = tilemap.center # Tilemap is always centered on PC
        self.speed = 100 / 1000.0 # in pixels/ms
        
    def update(self, delay):
        # Pretty dumb for now
        pcx,pcy = self.pc.rect.center
        x,y = self._rect.center
        dx = dy = 0
        
        dist = math.sqrt(math.pow(pcx - x,2) + math.pow(pcy - y,2))
        if dist < 100:
            self.target = self.pc
            self.attacking = True
            if pcx < x:
                dx = -1
            elif pcx > x:
                dx = 1
            else:
                dx = 0
                
            if pcy < y:
                dy = -1
            elif pcy > y:
                dy = 1
            else:
                dy = 0 
        else:
            self.target = None      
                 
        self.dx = dx
        self.dy = dy
        self.moving = dx or dy
        if self.moving:
            self.facing = self.DIRS_TO_FACING[ (dx,dy) ]
        
        NPC.update(self, delay)